"""
Utilities for Relations RestX
"""

import inspect

import flask_restx

from relations_restx.resource import ResourceError, ResourceIdentity, Resource, exceptions
from relations_restx.api import Api, OpenApi

def resources(module):
    """
    List all the resources in a module
    """

    return [
        cls[1]
        for cls in inspect.getmembers(
            module,
            lambda model: inspect.isclass(model)
            and issubclass(model, Resource)
        )
    ]

def ensure(module, models):
    """
    Creates the Resources for all models
    """

    exists = [resource.MODEL for resource in resources(module)]

    return [
        type(model.__name__, (Resource, ), {'MODEL': model})
        for model in models if model not in exists
    ]

def attach(restx, module, models):
    """
    Attach all Reources to a RestX
    """

    class Model(flask_restx.Resource):
        """
        Custom class for each call
        """

        MODELS = []

        def get(self):
            """
            List all models
            """
            return {"models": self.MODELS}

    restx.add_resource(Model, "/model")

    for resource in resources(module) + ensure(module, models):

        thy = resource.thy()

        Model.MODELS.append({
            "id": thy._model._id,
            "titles": thy._model._titles,
            "title": thy._model.TITLE,
            "singular": thy.SINGULAR,
            "plural": thy.PLURAL,
            "list": thy.LIST
        })

        if resource.__name__.lower() not in restx.endpoints:
            restx.add_resource(resource, *thy.endpoints())



"""
Resource module for Relations and Flask RestX
"""

# pylint: disable=not-callable


import flask
import flask_restx

import functools
import traceback
import werkzeug.exceptions

import opengui
import relations

def exceptions(endpoint):
    """
    Decorator that adds and handles a database session
    """

    @functools.wraps(endpoint)
    def wrap(*args, **kwargs):

        try:

            response = endpoint(*args, **kwargs)

        except werkzeug.exceptions.BadRequest as exception:

            response = {
                "message": exception.description
            }, 400

        except relations.ModelError as exception:

            message = str(exception)

            status_code = 404 if "none retrieved" in message else 500

            response = {
                "message": message,
            }, status_code

        except Exception as exception:

            response = {
                "message": str(exception),
                "traceback": traceback.format_exc()
            }, 500

        return response

    return wrap

class ResourceError(Exception):
    """
    Generic resource Error for easier tracing
    """

    def __init__(self, resource, message):

        self.resource = resource
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        """
        Might want to mention the resource and info about it
        """
        return f"{self.resource.__class__.__name__}: {self.message}"

class ResourceIdentity:
    """
    Intermediate static type class for constructing mode information with a full resource
    """

    MODEL = None
    SINGULAR = None
    PLURAL = None
    FIELDS = None
    LIST = None

    _model = None
    _fields = None

    @classmethod
    def thy(cls, self=None): # pylint: disable=too-many-branches
        """
        Base identity to be known without instantiating the class
        """

        # If self wasn't sent, we're just providing a shell of an instance

        if self is None:
            self = ResourceIdentity()
            self.__dict__.update(cls.__dict__)

        self._model = self.MODEL.thy()

        if self.SINGULAR is None:
            if hasattr(self._model, "SINGULAR") and self._model.SINGULAR is not None:
                self.SINGULAR = self._model.SINGULAR
            else:
                self.SINGULAR = self._model.NAME

        if self.PLURAL is None:
            if hasattr(self._model, "PLURAL") and self._model.PLURAL is not None:
                self.PLURAL = self._model.PLURAL
            else:
                self.PLURAL = f"{self.SINGULAR}s"

        if self.FIELDS is None:
            self.FIELDS = []

        self._fields = []
        fields = opengui.Fields(fields=self.FIELDS)

        for model_field in self._model._fields._order:

            form_field = {
                "name": model_field.name,
                "kind": model_field.kind.__name__
            }

            for attribute in ["options", "validation", "init", "inject"]:
                if getattr(model_field, attribute):
                    form_field[attribute] = getattr(model_field, attribute)

            if model_field.auto:
                form_field["readonly"] = True

            if model_field.default is not None:
                form_field["default"] = model_field.default() if callable(model_field.default) else model_field.default
                if isinstance(form_field["default"], set):
                    form_field["default"] = sorted(list(form_field["default"]))
            elif not model_field.auto and (not model_field.none or model_field.name in self._model._titles):
                form_field["required"] = True

            if model_field.name in fields.names:
                form_field.update(fields[model_field.name].to_dict())

            self._fields.append(form_field)

        if self.LIST is None:
            self.LIST = list(self._model._list)

        # Make sure all the list checks out

        for field in self.LIST:
            if field.split("__")[0] not in self._model._fields:
                raise ResourceError(self, f"cannot find field {field} from list")

        return self

    def endpoints(self):
        """
        Lists the endpoints this resource had
        """

        endpoints = [f"/{self.SINGULAR}"]

        if self._model.ID is not None:
            endpoints.append(f"/{self.SINGULAR}/<id>")

        return endpoints

class Resource(flask_restx.Resource, ResourceIdentity):
    """
    Base Model class for Relations Restful classes
    """

    def __init__(self, *args, **kwargs): # pylint: disable=super-init-not-called

        # Know thyself

        self.thy(self)

    @staticmethod
    def json():
        """
        Gets the current request JSON
        """

        try:
            return flask.request.json
        except: # pylint: disable=bare-except
            return {}

    @classmethod
    def criteria(cls, verify=False):
        """
        Gets criteria from the flask request
        """

        if verify and not flask.request.args and "filter" not in cls.json():
            raise werkzeug.exceptions.BadRequest("to confirm all, send a blank filter {}")

        criteria = {}

        if flask.request.args:
            criteria.update({
                name: value
                for name, value in flask.request.args.to_dict().items()
                if not name.startswith("limit") and name not in ["sort", "count"]
            })

        if "filter" in cls.json():
            criteria.update(flask.request.json["filter"])

        return criteria

    @classmethod
    def sort(cls):
        """
        Gets sort from the flask request
        """

        sort = []

        if flask.request.args and 'sort' in flask.request.args:
            sort.extend(flask.request.args['sort'].split(','))

        if "sort" in cls.json():
            sort.extend(flask.request.json['sort'])

        return sort

    @classmethod
    def limit(cls):
        """
        Gets limit from the flask request
        """

        limit = {}

        if flask.request.args:
            limit.update({
                name.split('__')[-1]: int(value)
                for name, value in flask.request.args.to_dict().items()
                if name.startswith("limit")
            })

        if "limit" in cls.json():
            limit.update({name: int(value) for name, value in flask.request.json["limit"].items()})

        return limit

    @classmethod
    def count(cls):
        """
        Gets count from the flask request
        """

        count = False

        if flask.request.args and 'count' in flask.request.args:
            count = flask.request.args['count']

        if "count" in cls.json():
            count = flask.request.json['count']

        if isinstance(count, (bool, int)):
            return count

        return count.lower() not in ["0", "no", "false"]

    def fields(self, likes, values, originals=None):
        """
        Apply options and titles to fields
        """

        fields = opengui.Fields(values=values, originals=originals, fields=self._fields)

        for field in fields:
            relation = self._model._ancestor(field.name)
            if relation is not None:
                like = {"like": likes[name] for name in likes if name == field.name}
                parent = relation.Parent.many(**like).limit()
                titles = parent.titles()

                field.content["format"] = titles.format
                field.content["overflow"] = parent.overflow

                value = field.value if field.value is not None else field.original

                if (not like and value is not None and value not in titles):
                    titles = relation.Parent.one(**{relation.parent_field: value}).titles()
                    field.content["overflow"] = True

                field.options = titles.ids
                field.content["titles"] = titles.titles

                field.content.update(like)

        return fields

    def formats(self, model):
        """
        Generate all the formats including parent lookups
        """

        formats = {}

        fields = opengui.Fields(fields=self._fields)

        for field in model._fields._order:
            relation = model._ancestor(field.name)
            if relation is not None:
                titles = relation.Parent.many(**{f"{relation.parent_field}__in": model[field.name]}).titles()
                formats[field.name] = {
                    "titles": titles.titles,
                    "format": titles.format
                }
            elif field.format is not None or "titles" in fields[field.name].content:
                formats[field.name] = {}
                if field.format is not None:
                    formats[field.name]["format"] = field.format
                if  "titles" in fields[field.name].content:
                    formats[field.name]["titles"] = fields[field.name].content["titles"]

        return formats

    @exceptions
    def options(self, id=None):
        """
        Generates form for inserts or updates of a single record
        """

        values = self.json().get(self.SINGULAR)
        likes = self.json().get("likes", {})

        if id is None:

            return self.fields(likes, values).to_dict(), 200

        originals = self.MODEL.one(**{self._model._id: id}).export()

        return self.fields(likes, values, originals).to_dict(), 200

    @exceptions
    def post(self):
        """
        Creates one or more models or complex
        """

        if "filter" in self.json():

            return self.get()

        if self.SINGULAR in self.json():

            return {self.SINGULAR: self.MODEL(**flask.request.json[self.SINGULAR]).create().export()}, 201

        if self.PLURAL in self.json():

            return {self.PLURAL: self.MODEL(flask.request.json[self.PLURAL]).create().export()}, 201

        raise werkzeug.exceptions.BadRequest(f"either {self.SINGULAR} or {self.PLURAL} required")

    @exceptions
    def get(self, id=None):
        """
        Retrieves one or more models
        """

        if id is not None:
            model = self.MODEL.one(**{self._model._id: id})
            return {self.SINGULAR: model.export(), "formats": self.formats(model)}

        models = self.MODEL.many(**self.criteria()).sort(*self.sort()).limit(**self.limit())

        if self.count():
            return {self.PLURAL: models.count(), "overflow": models.overflow}, 200

        return {self.PLURAL: models.export(), "overflow": models.overflow, "formats": self.formats(models)}, 200

    @exceptions
    def patch(self, id=None):
        """
        Updates one or more models models
        """

        if self.SINGULAR not in self.json() and self.PLURAL not in self.json():
            raise werkzeug.exceptions.BadRequest(f"either {self.SINGULAR} or {self.PLURAL} required")

        if id is not None:

            model = self.MODEL.one(**{self._model._id: id}).set(**flask.request.json[self.SINGULAR])

        elif self.SINGULAR in flask.request.json:

            model = self.MODEL.one(**self.criteria(True)).set(**flask.request.json[self.SINGULAR])

        elif self.PLURAL in flask.request.json:

            model = self.MODEL.many(**self.criteria(True)).set(**flask.request.json[self.PLURAL])

        return {"updated": model.update()}, 202

    @exceptions
    def delete(self, id=None):
        """
        Deletes one or more models models
        """

        if id is not None:

            model = self.MODEL.one(**{self._model._id: id})

        else:

            model = self.MODEL.many(**self.criteria(True))

        return {"deleted": model.delete()}, 202



"""
Module for overriding the base RestX API
"""

import collections
import flask_restx

from werkzeug.utils import cached_property


class OpenApi(flask_restx.Swagger):
    """
    Overrride Flask RestX Swagger
    """

    @staticmethod
    def relations_value(field): # pylint: disable=too-many-return-statements
        """
        Generates an example alues
        """

        if "default" in field:
            return field["default"]

        if field.get("options"):
            if field["kind"] == "set":
                return [field["options"][0]]
            return field["options"][0]

        if field["kind"] == "str":
            return ""

        if field["kind"] == "int":
            return 0

        if field["kind"] == "bool":
            return False

        if field["kind"] == "float":
            return 0.0

        return None

    @classmethod
    def relations_example(cls, thy, readonly=False):
        """
        Generates an example for a single record
        """

        example = {}

        for field in thy._fields:
            if readonly or not field.get("readonly"):
                example[field["name"]] = cls.relations_value(field)

        return example

    @staticmethod
    def relations_schemas(thy):
        """
        Generates specs from fields
        """

        record = {
            "type": "object",
            "properties": {},
        }

        required = []

        for field in thy._fields:

            property = {
                "type": field["kind"]
            }

            if field.get("readonly"):
                property["readOnly"] = True

            if field.get("required"):
                required.append(field["name"])

            record["properties"][field['name']] = property

        if required:
            record["required"] = required

        singular = {
            "type": "object",
            "properties": {
                thy.SINGULAR: {
                    "$ref": f"#/components/schemas/{thy._model.TITLE}"
                }
            }
        }

        plural = {
            "type": "object",
            "properties": {
                thy.PLURAL: {
                    "type": "array",
                    "items": {
                        "$ref": f"#/components/schemas/{thy._model.TITLE}"
                    }
                }
            }
        }

        filter = {
            "type": "object",
            "properties": {
                "filter": {
                    "$ref": f"#/components/schemas/{thy._model.TITLE}"
                }
            },
        }

        sort = {
            "type": "object",
            "properties": {
                "sort": {
                    "type": "array",
                    "description": "sort by these fields, prefix with + for ascending (default), - for descending",
                    "default": thy._model._order
                }
            }
        }

        limit = {
            "type": "object",
            "properties": {
                "limit": {
                    "type": "integer",
                    "description": f"limit the number of {thy.PLURAL}"
                },
                "limit__start": {
                    "type": "integer",
                    "description": f"limit the number of {thy.PLURAL} starting here"
                },
                "limit__per_page": {
                    "type": "integer",
                    "description": f"limit the number of {thy.PLURAL} by this page size (default {thy._model.CHUNK})"
                },
                "limit__page": {
                    "type": "integer",
                    "description": f"limit the number of {thy.PLURAL} and retrieve this page"
                }
            }
        }

        count = {
            "type": "object",
            "properties": {
                "count": {
                    "type": "boolean",
                    "description": f"return only the count of {thy.PLURAL} found"
                }
            }
        }

        return {
            thy._model.TITLE: record,
            thy.SINGULAR: singular,
            thy.PLURAL: plural,
            f"{thy.SINGULAR}_filter": filter,
            f"{thy.SINGULAR}_sort": sort,
            f"{thy.SINGULAR}_limit": limit,
            f"{thy.SINGULAR}_count": count
        }

    @classmethod
    def relations_create_options(cls, thy):
        """
        Generates create options operation
        """

        return {
            "tags": [thy._model.TITLE],
            "operationId": f"{thy.SINGULAR}_create_options",
            "summary": f"generates and validates fields to create one {thy.SINGULAR} or many {thy.PLURAL}",
            "description": f"To generate, send nothing. To validate, send a {thy.SINGULAR}.",
            "requestBody": {
                "content": {
                    "application/json": {
                        "schema": {
                            "$ref": f"#/components/schemas/{thy.SINGULAR}"
                        },
                        "examples": {
                            "generate": {
                                "value": {}
                            },
                            "validate": {
                                "value": {
                                    thy.SINGULAR: cls.relations_example(thy)
                                }
                            }
                        }
                    }
                }
            },
            "responses": {
                "200": {
                    "description": f"fields to create one {thy.SINGULAR} or many {thy.PLURAL} generated and validated",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": '#/components/schemas/Options'
                            }
                        }
                    }
                }
            }
        }

    @classmethod
    def relations_create_filter(cls, thy):
        """
        Generates create operation
        """

        return {
            "tags": [thy._model.TITLE],
            "operationId": f"{thy.SINGULAR}_create_search",
            "summary": f"creates one {thy.SINGULAR} or many {thy.PLURAL} or a complex retrieve",
            "description": f"To create one, send {thy.SINGULAR}. To create many, send {thy.PLURAL}. To retrieve send filter (sort, limit, count optional).",
            "requestBody": {
                "content": {
                    "application/json": {
                        "schema": {
                            "oneOf": [
                                {
                                    "$ref": f"#/components/schemas/{thy.SINGULAR}",
                                },
                                {
                                    "$ref": f"#/components/schemas/{thy.PLURAL}"
                                },
                                {
                                    "oneOf": [
                                        {
                                            "$ref": f"#/components/schemas/{thy.SINGULAR}_filter"
                                        }
                                    ],
                                    "anyOf": [
                                        {
                                            "$ref": f"#/components/schemas/{thy.SINGULAR}_sort"
                                        },
                                        {
                                            "$ref": f"#/components/schemas/{thy.SINGULAR}_limit"
                                        },
                                        {
                                            "$ref": f"#/components/schemas/{thy.SINGULAR}_count"
                                        }
                                    ]
                                }
                            ]
                        },
                        "examples": {
                            "create one": {
                                "value": {
                                    thy.SINGULAR: cls.relations_example(thy)
                                }
                            },
                            "create many": {
                                "value": {
                                    thy.PLURAL: [cls.relations_example(thy)]
                                }
                            },
                            "complex retrieve": {
                                "value": {
                                    "filter": cls.relations_example(thy),
                                    "sort": thy._model._order
                                }
                            },
                            "limit retrieve": {
                                "value": {
                                    "filter": cls.relations_example(thy),
                                    "sort": thy._model._order,
                                    "limit": {
                                        "limit": thy._model.CHUNK,
                                        "start": 0
                                    }
                                }
                            },
                            "paginate retrieve": {
                                "value": {
                                    "filter": cls.relations_example(thy),
                                    "sort": thy._model._order,
                                    "limit": {
                                        "page": 1,
                                        "per_page": thy._model.CHUNK
                                    }
                                }
                            },
                            "count retrieve": {
                                "value": {
                                    "filter": cls.relations_example(thy),
                                    "count": True
                                }
                            }
                        }
                    }
                }
            },
            "responses": {
                "200": {
                    "description": f"many {thy.PLURAL} retrieved",
                    "content": {
                        "application/json": {
                            "schema": {
                                "oneOf": [
                                    {
                                        "$ref": f"#/components/schemas/{thy.PLURAL}"
                                    },
                                    {
                                        "$ref": "#/components/schemas/Retrieved"
                                    },
                                ]
                            },
                            "examples": {
                                "list retrieve": {
                                    "value": {
                                        thy.PLURAL: [cls.relations_example(thy, readonly=True)],
                                        "overflow": False,
                                        "formats": {}
                                    }
                                },
                                "count retrieve": {
                                    "value": {
                                        "count": 1
                                    }
                                }
                            }
                        }
                    }
                },
                "201": {
                    "description": f"one {thy.SINGULAR} or many {thy.PLURAL} created",
                    "content": {
                        "application/json": {
                            "schema": {
                                "oneOf": [
                                    {
                                        "$ref": f"#/components/schemas/{thy.SINGULAR}",
                                        "description": "lol"
                                    },
                                    {
                                        "$ref": f"#/components/schemas/{thy.PLURAL}"
                                    }
                                ]
                            },
                            "examples": {
                                "create one": {
                                    "value": {
                                        thy.SINGULAR: cls.relations_example(thy, readonly=True)
                                    }
                                },
                                "create many": {
                                    "value": {
                                        thy.PLURAL: [cls.relations_example(thy, readonly=True)]
                                    }
                                }
                            }
                        }
                    }
                },
                "400": {
                    "description": "unable to create due to bad request"
                }
            }
        }

    @classmethod
    def relations_retrieve_many(cls, thy):
        """
        Generates reteieve many operation
        """

        return {
            "tags": [thy._model.TITLE],
            "operationId": f"{thy.SINGULAR}_retrieve_many",
            "summary": f"retrieves many {thy.PLURAL}",
            "parameters": [
                {
                    "in": "query",
                    "schema": {
                        "type": "object",
                        "properties": {
                            "dude": {
                                "type": "integer",
                                "example": 5,
                                "description": "lolwut"
                            }
                        },

                    },
                    "style": "form",
                    "explode": True,
                    "name": "params",
                    "examples": {
                        "retrieve": {
                            "value": {
                                **cls.relations_example(thy),
                                "sort": ",".join(thy._model._order)
                            }
                        },
                        "limit": {
                            "value": {
                                **cls.relations_example(thy),
                                "sort": ",".join(thy._model._order),
                                "limit": thy._model.CHUNK,
                                "limit__start": 0
                            }
                        },
                        "paginate": {
                            "value": {
                                **cls.relations_example(thy),
                                "sort": ",".join(thy._model._order),
                                "limit__page": 1,
                                "limit__per_page": thy._model.CHUNK
                            }
                        },
                        "count": {
                            "value": {
                                **cls.relations_example(thy),
                                "count": 1
                            }
                        }
                    }
                }
            ],
            "responses": {
                "200": {
                    "description": f"many {thy.PLURAL} retrieved",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": f"#/components/schemas/{thy.PLURAL}"
                            },
                            "examples": {
                                "list retrieve": {
                                    "value": {
                                        thy.PLURAL: [cls.relations_example(thy, readonly=True)],
                                        "overflow": False,
                                        "formats": {}
                                    }
                                },
                                "count retrieve": {
                                    "value": {
                                        "count": 1
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

    @classmethod
    def relations_update_many(cls, thy):
        """
        Generates update emany operation
        """

        return {
            "tags": [thy._model.TITLE],
            "operationId": f"{thy.SINGULAR}_update_many",
            "summary": f"updates many {thy.PLURAL}",
            "parameters": [
                {
                    "in": "query",
                    "schema": {
                        "type": "object"
                    },
                    "style": "form",
                    "explode": True,
                    "name": "params",
                    "examples": {
                        "filter through params": {
                            "value": {
                                **cls.relations_example(thy)
                            }
                        },
                        "filter through params limit": {
                            "value": {
                                **cls.relations_example(thy),
                                "sort": ",".join(thy._model._order),
                                "limit": thy._model.CHUNK,
                                "limit__start": 0
                            }
                        },
                        "filter through params paginate": {
                            "value": {
                                **cls.relations_example(thy),
                                "sort": ",".join(thy._model._order),
                                "limit__page": 1,
                                "limit__per_page": thy._model.CHUNK
                            }
                        },
                        "filter through body": {
                            "value": {}
                        }
                    }
                }
            ],
            "requestBody": {
                "content": {
                    "application/json": {
                        "schema": {
                            "anyOf": [
                                {
                                    "$ref": f"#/components/schemas/{thy.SINGULAR}"
                                },
                                {
                                    "$ref": f"#/components/schemas/{thy.SINGULAR}_filter"
                                }
                            ]
                        },
                        "examples": {
                            "filter through params": {
                                "value": {
                                    thy.PLURAL: cls.relations_example(thy)
                                }
                            },
                            "filter through body": {
                                "value": {
                                    "filter": cls.relations_example(thy),
                                    thy.PLURAL: cls.relations_example(thy)
                                }
                            },
                            "filter through body limit": {
                                "value": {
                                    "filter": cls.relations_example(thy),
                                    "sort": thy._model._order,
                                    "limit": {
                                        "limit": thy._model.CHUNK,
                                        "start": 0
                                    },
                                    thy.PLURAL: cls.relations_example(thy)
                                }
                            },
                            "filter through body paginate": {
                                "value": {
                                    "filter": cls.relations_example(thy),
                                    "sort": thy._model._order,
                                    "limit": {
                                        "page": 1,
                                        "per_page": thy._model.CHUNK
                                    },
                                    thy.PLURAL: cls.relations_example(thy)
                                }
                            },
                            "update all": {
                                "value": {
                                    "filter": {},
                                    thy.PLURAL: cls.relations_example(thy)
                                }
                            }
                        }
                    }
                }
            },
            "responses": {
                "202": {
                    "description": f"many {thy.PLURAL} updated",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/Updated"
                            }
                        }
                    }
                },
                "400": {
                    "description": "unable to update due to bad request"
                }
            }
        }

    @classmethod
    def relations_delete_many(cls, thy):
        """
        Generates delete many operation
        """

        return {
            "tags": [thy._model.TITLE],
            "operationId": f"{thy.SINGULAR}_delete_many",
            "summary": f"deletes many {thy.PLURAL}",
            "parameters": [
                {
                    "in": "query",
                    "schema": {
                        "type": "object"
                    },
                    "style": "form",
                    "explode": True,
                    "name": "params",
                    "examples": {
                        "filter through params": {
                            "value": {
                                **cls.relations_example(thy)
                            }
                        },
                        "filter through params limit": {
                            "value": {
                                **cls.relations_example(thy),
                                "sort": ",".join(thy._model._order),
                                "limit": thy._model.CHUNK,
                                "limit__start": 0
                            }
                        },
                        "filter through params paginate": {
                            "value": {
                                **cls.relations_example(thy),
                                "sort": ",".join(thy._model._order),
                                "limit__page": 1,
                                "limit__per_page": thy._model.CHUNK
                            }
                        },
                        "filter through body": {
                            "value": {}
                        }
                    }
                }
            ],
            "requestBody": {
                "content": {
                    "application/json": {
                        "schema": {
                            "anyOf": [
                                {
                                    "$ref": f"#/components/schemas/{thy.SINGULAR}_filter"
                                }
                            ]
                        },
                        "examples": {
                            "filter through params": {
                                "value": {}
                            },
                            "filter through body": {
                                "value": {
                                    "filter": cls.relations_example(thy)
                                }
                            },
                            "filter through body limit": {
                                "value": {
                                    "filter": cls.relations_example(thy),
                                    "sort": thy._model._order,
                                    "limit": {
                                        "limit": thy._model.CHUNK,
                                        "start": 0
                                    }
                                }
                            },
                            "filter through body paginate": {
                                "value": {
                                    "filter": cls.relations_example(thy),
                                    "sort": thy._model._order,
                                    "limit": {
                                        "page": 1,
                                        "per_page": thy._model.CHUNK
                                    }
                                }
                            },
                            "delete all": {
                                "value": {
                                    "filter": {}
                                }
                            }
                        }
                    }
                }
            },
            "responses": {
                "202": {
                    "description": f"many {thy.PLURAL} deleted",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/Deleted"
                            }
                        }
                    }
                },
                "400": {
                    "description": "unable to updeletedate due to bad request"
                }
            }
        }

    @classmethod
    def relations_update_options(cls, thy):
        """
        Generates update options operation
        """

        return {
            "tags": [thy._model.TITLE],
            "operationId": f"{thy.SINGULAR}_update_options",
            "summary": f"generates and validates fields to update one {thy.SINGULAR}",
            "requestBody": {
                "content": {
                    "application/json": {
                        "schema": {
                            "$ref": f"#/components/schemas/{thy.SINGULAR}"
                        },
                        "examples": {
                            "generate": {
                                "value": {}
                            },
                            "validate": {
                                "value": {
                                    thy.SINGULAR: cls.relations_example(thy)
                                }
                            }
                        }
                    }
                }
            },
            "responses": {
                "200": {
                    "description": f"fields to update one {thy.SINGULAR} generated and validated",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": '#/components/schemas/Options'
                            }
                        }
                    }
                },
                "404": {
                    "description": f"{thy.SINGULAR} not found"
                }
            }
        }

    @classmethod
    def relations_retrieve_one(cls, thy):
        """
        Generates retrieve one operation
        """

        return {
            "tags": [thy._model.TITLE],
            "operationId": f"{thy.SINGULAR}_retrieve_one",
            "summary": f"retrieves one {thy.SINGULAR}",
            "responses": {
                "200": {
                    "description": f"one {thy.SINGULAR} retrieved",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": f"#/components/schemas/{thy.SINGULAR}"
                            },
                            "examples": {
                                "retrieve": {
                                    "value": {
                                        thy.SINGULAR: cls.relations_example(thy, readonly=True),
                                        "overflow": False,
                                        "formats": {}
                                    }
                                }
                            }
                        }
                    }
                },
                "404": {
                    "description": f"{thy.SINGULAR} not found"
                }
            }
        }

    @classmethod
    def relations_update_one(cls, thy):
        """
        Generates update one operation
        """

        return {
            "tags": [thy._model.TITLE],
            "operationId": f"{thy.SINGULAR}_update_one",
            "summary": f"updates one {thy.SINGULAR}",
            "requestBody": {
                "content": {
                    "application/json": {
                        "schema": {
                            "$ref": f"#/components/schemas/{thy.SINGULAR}"
                        },
                        "examples": {
                            "update": {
                                "value": {
                                    thy.SINGULAR: cls.relations_example(thy)
                                }
                            }
                        }
                    }
                }
            },
            "responses": {
                "202": {
                    "description": f"one {thy.SINGULAR} updated",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/Updated"
                            }
                        }
                    }
                },
                "400": {
                    "description": "unable to update due to bad request"
                },
                "404": {
                    "description": f"{thy.SINGULAR} not found"
                }
            }
        }

    @classmethod
    def relations_delete_one(cls, thy):
        """
        Generates delete one operation
        """

        return {
            "tags": [thy._model.TITLE],
            "operationId": f"{thy.SINGULAR}_delete_one",
            "summary": f"deletes one {thy.SINGULAR}",
            "responses": {
                "202": {
                    "description": f"one {thy.SINGULAR} deleted",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/"
                            }
                        }
                    }
                },
                "404": {
                    "description": f"{thy.SINGULAR} not found"
                }
            }
        }

    def relations_operations(self, specs, ns, urls, thy): # pylint: disable=too-many-branches
        """
        Generates operations for all methods
        """

        for url in self.api.ns_urls(ns, urls):

            path = flask_restx.swagger.extract_path(url)

            methods = specs["paths"][path]
            specs["paths"][path] = collections.OrderedDict()

            for method in ["options", "post", "get", "patch", "delete"]:

                if "{" not in path:
                    if method == "options":
                        specs["paths"][path][method] = {**methods[method], **self.relations_create_options(thy)}
                    elif method == "post":
                        specs["paths"][path][method] = {**methods[method], **self.relations_create_filter(thy)}
                    elif method == "get":
                        specs["paths"][path][method] = {**methods[method], **self.relations_retrieve_many(thy)}
                    elif method == "patch":
                        specs["paths"][path][method] = {**methods[method], **self.relations_update_many(thy)}
                    elif method == "delete":
                        specs["paths"][path][method] = {**methods[method], **self.relations_delete_many(thy)}
                else:
                    if method == "options":
                        specs["paths"][path][method] = {**methods[method], **self.relations_update_options(thy)}
                    elif method == "post":
                        continue
                    elif method == "get":
                        specs["paths"][path][method] = {**methods[method], **self.relations_retrieve_one(thy)}
                    elif method == "patch":
                        specs["paths"][path][method] = {**methods[method], **self.relations_update_one(thy)}
                    elif method == "delete":
                        specs["paths"][path][method] = {**methods[method], **self.relations_delete_one(thy)}

    def relations_resource(self, specs, ns, resource, urls):
        """
        Overrides OpenApi specs for a Relations Resource
        """

        thy = resource.thy()

        specs["tags"].append({
            "name": thy._model.TITLE
        })

        specs["components"]["schemas"].update(self.relations_schemas(thy))

        self.relations_operations(specs, ns, urls, thy)

    def as_dict(self):
        """
        Overides swagger dict to make it OpenAPI
        """

        specs = super().as_dict()

        del specs["swagger"]
        specs["openapi"] = "3.0.3"

        specs.setdefault("components", {})
        specs["components"].setdefault("schemas", {})

        specs["components"]["schemas"].update({
            "Field": {
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string",
                        "description": "name of the field"
                    },
                    "value": {
                        "description": "the current value of the field"
                    },
                    "original": {
                        "description": "the original value of the field"
                    },
                    "default": {
                        "description": "the default value of the field"
                    },
                    "options": {
                        "type": "array",
                        "items": {},
                        "description": "array of options to select from"
                    },
                    "required": {
                        "type": "boolean",
                        "description": "whether the field is required"
                    },
                    "multi": {
                        "type": "boolean",
                        "description": "whether multiple options can be selected"
                    },
                    "trigger": {
                        "type": "boolean",
                        "description": "whether to reload when this field changes"
                    },
                    "readonly": {
                        "type": "boolean",
                        "description": "whether the field is readonly"
                    },
                    "validation": {
                        "description": "how to validate this field"
                    },
                    "content": {
                        "type": "object",
                        "description": "used for any other data, like titles"
                    },
                    "errors": {
                        "type": "array",
                        "description": "the original value of the field"
                    }
                }
            },
            "Options": {
                "type": "object",
                "properties": {
                    "fields": {
                        "type": "array",
                        "items": {
                            "$ref": "#/components/schemas/Field"
                        }
                    },
                    "errors": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    }
                }
            },
            "Retrieved": {
                "type": "object",
                "properties": {
                    "overflow": {
                        "type": "boolean",
                        "description": "whether more could have been retrieved"
                    },
                    "format": {
                        "type": "object",
                        "description": "Formatting information for fields, like titles"
                    }
                }
            },
            "Counted": {
                "type": "object",
                "properties": {
                    "count": {
                        "type": "integer",
                        "description": "count of those retrieved"
                    }
                }
            },
            "Updated": {
                "type": "object",
                "properties": {
                    "updated": {
                        "type": "integer",
                        "description": "count of those updated"
                    }
                }
            },
            "Deleted": {
                "type": "object",
                "properties": {
                    "deleted": {
                        "type": "integer",
                        "description": "count of those deleted"
                    }
                }
            }
        })

        for ns in self.api.namespaces:
            for resource, urls, _, _ in ns.resources:
                if hasattr(resource, "thy"):
                    self.relations_resource(specs, ns, resource, urls)

        return specs


class Api(flask_restx.Api):
    """
    Overrride Flask RestX API
    """

    @cached_property
    def __schema__(self):
        """
        The Swagger specifications/schema for this API

        :returns dict: the schema as a serializable dict
        """
        if not self._schema:
            try:
                self._schema = OpenApi(self).as_dict()
            except Exception: # pragma: no cover
                # Log the source exception for debugging purpose
                # and return an error message
                msg = "Unable to render schema"
                flask_restx.api.log.exception(msg)  # This will provide a full traceback
                return {"error": msg}
        return self._schema
